import { toArray } from './collectors/toArray';
import { IntermediateStage } from './stages';
/**
 * A reusable pipeline that lazily evaluates the data processing operations which are referred to as stages.
 * These stages are only evaluated when a terminal stage is triggered by calling
 * {@link collect()}, passing in an implementation of a terminal operation or by calling or {@link toArray()}.
 *
 * Each lazy pipeline is a stage itself that acts as the source which in turn pipes
 * each element in the underlying data source to its downstream stages.
 *
 * A notable feature of this pipeline is that it is reusable, so you can construct your pipeline however
 * you'd like and reuse it in different places with no worries about data corruptions by calling {@link reset()}
 * before triggering a terminal stage on the pipeline, otherwise, an error will occur.
 *
 * @template IN The type parameter of each element in the pipeline.
 */
export class LazyPipeline extends IntermediateStage {
    /**
     * The intermediate stages that were added to this pipeline so far
     */
    _stages = [this];
    /**
     * The data source to operate upon
     */
    _source;
    _isTerminated = false;
    _isFrozen = false;
    _terminalStage;
    constructor(source) {
        super();
        this._source = Array.from(source);
        this._watchForPipelineEvent = this._watchForPipelineEvent.bind(this);
    }
    static from(iterable) {
        return new LazyPipeline(iterable);
    }
    add(...stages) {
        if (this._isFrozen) {
            throw new Error('[LazyPipeline] Pipeline has been frozen, please call unfreeze() before adding new intermediate stages');
        }
        this._attachNewStages(stages);
        if (this._stages.length === 1) {
            this.pipeTo(stages[0]);
        }
        this._stages.push(...stages);
        return this;
    }
    _attachNewStages(stages) {
        const totalNewStages = stages.length;
        for (let index = 0; index < totalNewStages; index++) {
            const currentStage = stages[index];
            const downstreamStage = stages[index + 1];
            if (downstreamStage) {
                currentStage.pipeTo(downstreamStage);
            }
            currentStage.subscribe(this._watchForPipelineEvent);
        }
    }
    _watchForPipelineEvent(event, stage) {
        switch (event) {
            case "TERMINATE_EARLY" /* PipelineEvent.TERMINATE_EARLY */:
                this._isTerminated = true;
                break;
            case "DETACH_STAGE" /* PipelineEvent.DETACH_STAGE */:
                if (stage instanceof IntermediateStage) {
                    const indexOfStageToDetach = this._stages.indexOf(stage);
                    const stageBefore = this._stages[indexOfStageToDetach - 1];
                    const stageAfter = this._stages[indexOfStageToDetach + 1];
                    stage.detach();
                    stageBefore?.pipeTo(stageAfter || this._terminalStage);
                }
                else {
                    throw new Error(`[LazyPipeline] A non-intermediate stage was detached, detached staged was ${JSON.stringify(stage)}`);
                }
        }
    }
    collect(collector) {
        if (this._isTerminated) {
            throw new Error('Pipeline has already been consumed, please call reset() before attempting to consume this pipeline again');
        }
        const lastStage = this._stages[this._stages.length - 1];
        lastStage.pipeTo(collector);
        this._terminalStage = collector;
        collector.subscribe(this._watchForPipelineEvent);
        const totalElementCount = this._source.length;
        for (let index = 0; index < totalElementCount && !this._isTerminated; index++) {
            this.consume(this._source[index], index !== totalElementCount - 1);
        }
        this._isFrozen = true;
        this._isTerminated = true;
        return collector.get();
    }
    toArray() {
        return this.collect(toArray());
    }
    reset() {
        this._isTerminated = false;
        this._downstream.reset();
        this._reconnectAllStages();
    }
    _reconnectAllStages() {
        const totalStages = this._stages.length;
        for (let index = 0; index < totalStages - 1; index++) {
            const currentStage = this._stages[index];
            const downstreamStage = this._stages[index + 1];
            currentStage.pipeTo(downstreamStage);
        }
    }
    readFrom(newSource) {
        this._source = Array.from(newSource);
    }
    freeze() {
        this._isFrozen = true;
    }
    unfreeze() {
        this._isFrozen = false;
    }
    consume(element, hasMoreDataUpstream) {
        this._downstream.consume(element, hasMoreDataUpstream);
    }
}
//# sourceMappingURL=LazyPipeline.js.map